@@ -1,8 +1,8 @@
 /* ------------------------------------------------------------------------- */
 /*   "expressp" :  The expression parser                                     */
 /*                                                                           */
-/*   Part of Inform 6.30                                                     */
-/*   copyright (c) Graham Nelson 1993 - 2004                                 */
+/*   Part of Inform 6.34                                                     */
+/*   copyright (c) Graham Nelson 1993 - 2017                                 */
 /*                                                                           */
 /* ------------------------------------------------------------------------- */
 
@@ -47,12 +47,9 @@ static void make_lexical_interface_tables(void)
 
 static token_data current_token, previous_token, heldback_token;
 
-static void function_call_triggered(void);
-
 static int comma_allowed, arrow_allowed, superclass_allowed,
            bare_prop_allowed,
            array_init_ambiguity, action_ambiguity,
-
            etoken_count, inserting_token, bracket_level;
 
 extern int *variable_usage;
@@ -60,7 +57,7 @@ extern int *variable_usage;
 int system_function_usage[32];
 
 static int get_next_etoken(void)
-{   int v, symbol, mark_symbol_as_used = FALSE,
+{   int v, symbol = 0, mark_symbol_as_used = FALSE,
         initial_bracket_level = bracket_level;
 
     etoken_count++;
@@ -131,6 +128,14 @@ but not used as a value:", unicode);
             current_token.symflags = sflags[symbol];
             switch(stypes[symbol])
             {   case ROUTINE_T:
+                    /* Replaced functions must always be backpatched
+                       because there could be another definition coming. */
+                    if (sflags[symbol] & REPLACE_SFLAG)
+                    {   current_token.marker = SYMBOL_MV;
+                        if (module_switch) import_symbol(symbol);
+                        v = symbol;
+                        break;
+                    }
                     current_token.marker = IROUTINE_MV;
                     break;
                 case GLOBAL_VARIABLE_T:
@@ -248,7 +253,6 @@ but not used as a value:", unicode);
                         heldback_token = current_token;
                         current_token.text = "<call>";
                         bracket_level--;
-                        function_call_triggered();
                     }
                     break;
 
@@ -299,6 +303,40 @@ but not used as a value:", unicode);
                     token_text = current_token.text + 3;
                     goto ActionUsedAsConstant;
 
+                case HASHGDOLLAR_SEP:
+
+                /* This form generates the position of a global variable
+                   in the global variables array. So Glob is the same as
+                   #globals_array --> #g$Glob                                */
+
+                    current_token.text += 3;
+                    current_token.type = SYMBOL_TT;
+                    symbol = symbol_index(current_token.text, -1);
+                    if (stypes[symbol] != GLOBAL_VARIABLE_T) {
+                        ebf_error(
+                        "global variable name after '#g$'",
+                        current_token.text);
+                        current_token.value = 0;
+                        current_token.type = SMALL_NUMBER_TT;
+                        current_token.marker = 0;
+                        break;
+                    }
+                    mark_symbol_as_used = TRUE;
+                    current_token.value = svals[symbol] - MAX_LOCAL_VARIABLES;
+                    current_token.marker = 0;
+                    if (!glulx_mode) {
+                        if (current_token.value >= 0x100)
+                            current_token.type = LARGE_NUMBER_TT;
+                        else current_token.type = SMALL_NUMBER_TT;
+                    }
+                    else {
+                        if (current_token.value >= 0x8000
+                          || current_token.value < -0x8000) 
+                            current_token.type = LARGE_NUMBER_TT;
+                        else current_token.type = SMALL_NUMBER_TT;
+                    }
+                    break;
+
                 case HASHNDOLLAR_SEP:
 
                 /* This form is still needed for constants like #n$a (the
@@ -338,7 +376,7 @@ but not used as a value:", unicode);
                     system_constants.enabled = FALSE;
                     if (token_type != SYSTEM_CONSTANT_TT)
                     {   ebf_error(
-                        "'r$', 'n$' or internal Inform constant name after '#'",
+                        "'r$', 'n$', 'g$' or internal Inform constant name after '#'",
                         token_text);
                         break;
                     }
@@ -506,6 +544,52 @@ static int find_prec(token_data a, token_data b)
 
 /* --- Converting token to operand ----------------------------------------- */
 
+/* Must match the switch statement below */
+int z_system_constant_list[] =
+    { adjectives_table_SC,
+      actions_table_SC,
+      classes_table_SC,
+      identifiers_table_SC,
+      preactions_table_SC,
+      largest_object_SC,
+      strings_offset_SC,
+      code_offset_SC,
+      actual_largest_object_SC,
+      static_memory_offset_SC,
+      array_names_offset_SC,
+      readable_memory_offset_SC,
+      cpv__start_SC,
+      cpv__end_SC,
+      ipv__start_SC,
+      ipv__end_SC,
+      array__start_SC,
+      array__end_SC,
+      highest_attribute_number_SC,
+      attribute_names_array_SC,
+      highest_property_number_SC,
+      property_names_array_SC,
+      highest_action_number_SC,
+      action_names_array_SC,
+      highest_fake_action_number_SC,
+      fake_action_names_array_SC,
+      highest_routine_number_SC,
+      routine_names_array_SC,
+      routines_array_SC,
+      routine_flags_array_SC,
+      highest_global_number_SC,
+      global_names_array_SC,
+      globals_array_SC,
+      global_flags_array_SC,
+      highest_array_number_SC,
+      array_names_array_SC,
+      array_flags_array_SC,
+      highest_constant_number_SC,
+      constant_names_array_SC,
+      highest_class_number_SC,
+      class_objects_array_SC,
+      highest_object_number_SC,
+      -1 };
+
 static int32 value_of_system_constant_z(int t)
 {   switch(t)
     {   case adjectives_table_SC:
@@ -605,6 +689,22 @@ static int32 value_of_system_constant_z(int t)
     return(0);
 }
 
+/* Must match the switch statement below */
+int glulx_system_constant_list[] =
+    { classes_table_SC,
+      identifiers_table_SC,
+      array_names_offset_SC,
+      cpv__start_SC,
+      cpv__end_SC,
+      dictionary_table_SC,
+      dynam_string_table_SC,
+      grammar_table_SC,
+      actions_table_SC,
+      globals_array_SC,
+      highest_class_number_SC,
+      highest_object_number_SC,
+      -1 };
+
 static int32 value_of_system_constant_g(int t)
 { 
   switch (t) {
@@ -626,6 +726,12 @@ static int32 value_of_system_constant_g(int t)
     return grammar_table_offset;
   case actions_table_SC:
     return actions_offset;
+  case globals_array_SC:
+    return variables_offset;
+  case highest_class_number_SC:
+    return no_classes-1;
+  case highest_object_number_SC:
+    return no_objects-1;
   }
 
   error_named("System constant not implemented in Glulx",
@@ -780,22 +886,43 @@ static int evaluate_term(token_data t, assembly_operand *o)
                  o->type = CONSTANT_OT;
                  switch(t.value)
                  {
+                 /* The three dict_par flags point at the lower byte
+                    of the flag field, because the library is written
+                    to expect one-byte fields, even though the compiler
+                    generates a dictionary with room for two. */
                  case dict_par1_SC:
                      o->type = BYTECONSTANT_OT;
                      o->marker = 0;
-                     v = DICT_WORD_SIZE+2;
+                     v = DICT_ENTRY_FLAG_POS+1;
                      break;
                  case dict_par2_SC:
                      o->type = BYTECONSTANT_OT;
                      o->marker = 0;
-                     v = DICT_WORD_SIZE+4;
+                     v = DICT_ENTRY_FLAG_POS+3;
                      break;
                  case dict_par3_SC:
                      o->type = BYTECONSTANT_OT;
                      o->marker = 0;
-                     v = DICT_WORD_SIZE+6;
+                     v = DICT_ENTRY_FLAG_POS+5;
                      break;
 
+                 case lowest_attribute_number_SC:
+                 case lowest_action_number_SC:
+                 case lowest_routine_number_SC:
+                 case lowest_array_number_SC:
+                 case lowest_constant_number_SC:
+                 case lowest_class_number_SC:
+                     o->type = BYTECONSTANT_OT;
+                     o->marker = 0;
+                     v = 0;
+                     break;
+                 case lowest_object_number_SC:
+                 case lowest_property_number_SC:
+                     o->type = BYTECONSTANT_OT;
+                     o->marker = 0;
+                     v = 1;
+                     break;
+ 
                  /* ###fix: need to fill more of these in! */
 
                  default:
@@ -822,89 +949,152 @@ extern void clear_expression_space(void)
 
 static assembly_operand *emitter_stack;
 static int *emitter_markers;
+static int *emitter_bracket_counts;
 
-#define FVALUE_MARKER 1
-#define ORVALUE_MARKER 2
+#define FUNCTION_VALUE_MARKER 1
+#define ARGUMENT_VALUE_MARKER 2
+#define OR_VALUE_MARKER 3
 
-static int emitter_sp, next_marker = 0;
-
-static void function_call_triggered(void)
-{   next_marker = FVALUE_MARKER;
-    if (expr_trace_level >= 3) printf("Function call triggered\n");
-}
+static int emitter_sp;
 
 static int is_property_t(int symbol_type)
 {   return ((symbol_type == PROPERTY_T) || (symbol_type == INDIVIDUAL_PROPERTY_T));
 }
 
+static void mark_top_of_emitter_stack(int marker, token_data t)
+{   if (emitter_sp < 1)
+    {   compiler_error("SR error: Attempt to add a marker to the top of an empty emitter stack");
+        return;
+    }
+    if (expr_trace_level >= 2)
+    {   printf("Marking top of emitter stack (which is ");
+        print_operand(emitter_stack[emitter_sp-1]);
+        printf(") as ");
+        switch(marker)
+        {
+            case FUNCTION_VALUE_MARKER:
+                printf("FUNCTION");
+                break;
+            case ARGUMENT_VALUE_MARKER:
+                printf("ARGUMENT");
+                break;
+            case OR_VALUE_MARKER:
+                printf("OR_VALUE");
+                break;
+            default:
+                printf("UNKNOWN");
+                break;
+        }
+        printf("\n");
+    }
+    if (emitter_markers[emitter_sp-1])
+    {   if (marker == ARGUMENT_VALUE_MARKER)
+        {
+            warning("Ignoring spurious leading comma");
+            return;
+        }
+        error_named("Missing operand for", t.text);
+        if (emitter_sp == MAX_EXPRESSION_NODES)
+            memoryerror("MAX_EXPRESSION_NODES", MAX_EXPRESSION_NODES);
+        emitter_markers[emitter_sp] = 0;
+        emitter_bracket_counts[emitter_sp] = 0;
+        emitter_stack[emitter_sp] = zero_operand;
+        emitter_sp++;
+    }
+    emitter_markers[emitter_sp-1] = marker;
+}
+
+static void add_bracket_layer_to_emitter_stack(int depth)
+{   /* There's no point in tracking bracket layers that don't fence off any values. */
+    if (emitter_sp < depth + 1) return;
+    if (expr_trace_level >= 2)
+        printf("Adding bracket layer\n");
+    ++emitter_bracket_counts[emitter_sp-depth-1];
+}
+
+static void remove_bracket_layer_from_emitter_stack()
+{   /* Bracket layers that don't fence off any values will not have been tracked. */
+    if (emitter_sp < 2) return;
+    if (expr_trace_level >= 2)
+        printf("Removing bracket layer\n");
+    if (emitter_bracket_counts[emitter_sp-2] <= 0)
+    {   compiler_error("SR error: Attempt to remove a nonexistent bracket layer from the emitter stack");
+        return;
+    }
+    --emitter_bracket_counts[emitter_sp-2];
+}
+
 static void emit_token(token_data t)
-{   assembly_operand o1, o2; int arity, stack_size, i, mark_result = 0;
+{   assembly_operand o1, o2; int arity, stack_size, i;
     int op_node_number, operand_node_number, previous_node_number;
-    int32 x;
+    int32 x = 0;
 
     if (expr_trace_level >= 2)
     {   printf("Output: %-19s%21s ", t.text, "");
         for (i=0; i<emitter_sp; i++)
-        {   if (emitter_markers[i] == FVALUE_MARKER) printf(" FVALUE:");
-            if (emitter_markers[i] == ORVALUE_MARKER) printf(" ORVALUE:");
-            print_operand(emitter_stack[i]); printf(" ");
+        {   print_operand(emitter_stack[i]); printf(" ");
+            if (emitter_markers[i] == FUNCTION_VALUE_MARKER) printf(":FUNCTION ");
+            if (emitter_markers[i] == ARGUMENT_VALUE_MARKER) printf(":ARGUMENT ");
+            if (emitter_markers[i] == OR_VALUE_MARKER) printf(":OR ");
+            if (emitter_bracket_counts[i]) printf(":BRACKETS(%d) ", emitter_bracket_counts[i]);
         }
         printf("\n");
     }
 
     if (t.type == SUBOPEN_TT) return;
 
+    stack_size = 0;
+    while ((stack_size < emitter_sp) &&
+           !emitter_markers[emitter_sp-stack_size-1] &&
+           !emitter_bracket_counts[emitter_sp-stack_size-1])
+        stack_size++;
+
     if (t.type == SUBCLOSE_TT)
-    {   if (emitter_sp == 0)
-        {   error("No expression between brackets '(' and ')'");
-            emitter_markers[0] = 0;
-            emitter_stack[0] = zero_operand;
-            emitter_sp = 1;
-        }
-        if (next_marker == FVALUE_MARKER)
-        {   emitter_markers[emitter_sp-1] = next_marker; next_marker=0;
+    {   if (stack_size < emitter_sp && emitter_bracket_counts[emitter_sp-stack_size-1])
+        {   if (stack_size == 0)
+            {   error("No expression between brackets '(' and ')'");
+                emitter_stack[emitter_sp] = zero_operand;
+                emitter_markers[emitter_sp] = 0;
+                emitter_bracket_counts[emitter_sp] = 0;
+                ++emitter_sp;
+            }
+            else if (stack_size < 1)
+                compiler_error("SR error: emitter stack empty in subexpression");
+            else if (stack_size > 1)
+                compiler_error("SR error: emitter stack overfull in subexpression");
+            remove_bracket_layer_from_emitter_stack();
         }
         return;
     }
 
     if (t.type != OP_TT)
-    {   emitter_markers[emitter_sp] = next_marker;
-
-        if (t.type == SYSFUN_TT)
-        {   if (next_marker != FVALUE_MARKER)
-            {   error_named("System function name used as a value:", t.text);
-                t.type = LARGE_NUMBER_TT;
-                t.value = 0;
-                t.marker = 0;
-            }
-        }
-        next_marker=0;
+    {   emitter_markers[emitter_sp] = 0;
+        emitter_bracket_counts[emitter_sp] = 0;
 
         if (emitter_sp == MAX_EXPRESSION_NODES)
-            memoryerror("MAX_EXPRESSION_NODES1", MAX_EXPRESSION_NODES);
+            memoryerror("MAX_EXPRESSION_NODES", MAX_EXPRESSION_NODES);
         if (!evaluate_term(t, &(emitter_stack[emitter_sp++])))
             compiler_error_named("Emit token error:", t.text);
         return;
     }
 
-    stack_size = 0;
-    while ((emitter_sp > stack_size)
-           && (emitter_markers[emitter_sp-stack_size-1] != FVALUE_MARKER))
-        stack_size++;
-
-    if ((emitter_sp > 1)&&(stack_size == 0)
-        &&(operators[t.value].precedence >= 12))
-    {   emitter_markers[emitter_sp - 1] = 0;
-        stack_size = 2;
-        mark_result = FVALUE_MARKER;
-    }
-
-    if ((t.value == COMMA_OP) && (emitter_sp != stack_size)) return;
-    if (t.value == OR_OP)
-    {   emitter_markers[emitter_sp-1] = ORVALUE_MARKER;
+    /* A comma is argument-separating if it follows an argument (or a function
+       call, since we ignore spurious leading commas in function argument lists)
+       with no intervening brackets.  Function calls are variadic, so we don't
+       apply argument-separating commas. */
+    if (t.value == COMMA_OP &&
+        stack_size < emitter_sp &&
+        (emitter_markers[emitter_sp-stack_size-1] == ARGUMENT_VALUE_MARKER ||
+         emitter_markers[emitter_sp-stack_size-1] == FUNCTION_VALUE_MARKER) &&
+        !emitter_bracket_counts[emitter_sp-stack_size-1])
+    {   if (expr_trace_level >= 2)
+            printf("Treating comma as argument-separating\n");
         return;
     }
 
+    if (t.value == OR_OP)
+        return;
+
     arity = 1;
     if (t.value == FCALL_OP)
     {   if (expr_trace_level >= 3)
@@ -912,8 +1102,27 @@ static void emit_token(token_data t)
             for (x=0; x<emitter_sp; x++) printf("%d ", emitter_markers[x]);
             printf("\n");
         }
-        while (emitter_markers[emitter_sp-arity] != FVALUE_MARKER)
-            arity++;
+        if (emitter_markers[emitter_sp-1] == ARGUMENT_VALUE_MARKER)
+            warning("Ignoring spurious trailing comma");
+        while (emitter_markers[emitter_sp-arity] != FUNCTION_VALUE_MARKER)
+        {
+            if ((glulx_mode &&
+                 emitter_stack[emitter_sp-arity].type == SYSFUN_OT) ||
+                (!glulx_mode &&
+                 emitter_stack[emitter_sp-arity].type == VARIABLE_OT &&
+                 emitter_stack[emitter_sp-arity].value >= 256 &&
+                 emitter_stack[emitter_sp-arity].value < 288))
+            {   int index = emitter_stack[emitter_sp-arity].value;
+                if(!glulx_mode)
+                    index -= 256;
+                if(index >= 0 && index < NUMBER_SYSTEM_FUNCTIONS)
+                    error_named("System function name used as a value:", system_functions.keywords[index]);
+                else
+                    compiler_error("Found unnamed system function used as a value");
+                emitter_stack[emitter_sp-arity] = zero_operand;
+            }
+            ++arity;
+        }
     }
     else
     {   arity = 1;
@@ -921,16 +1130,24 @@ static void emit_token(token_data t)
 
         if (operators[t.value].precedence == 3)
         {   arity = 2;
-            for (x = emitter_sp-1; emitter_markers[x]==ORVALUE_MARKER; x--)
-                arity++;
+            x = emitter_sp-1;
+            if(!emitter_markers[x] && !emitter_bracket_counts[x])
+            {   for (--x; emitter_markers[x] == OR_VALUE_MARKER && !emitter_bracket_counts[x]; --x)
+                {   ++arity;
+                    ++stack_size;
+                }
+                for (;x >= 0 && !emitter_markers[x] && !emitter_bracket_counts[x]; --x)
+                    ++stack_size;
+            }
         }
 
         if (arity > stack_size)
         {   error_named("Missing operand for", t.text);
             while (arity > stack_size)
             {   if (emitter_sp == MAX_EXPRESSION_NODES)
-                    memoryerror("MAX_EXPRESSION_NODES2", MAX_EXPRESSION_NODES);
+                    memoryerror("MAX_EXPRESSION_NODES", MAX_EXPRESSION_NODES);
                 emitter_markers[emitter_sp] = 0;
+                emitter_bracket_counts[emitter_sp] = 0;
                 emitter_stack[emitter_sp] = zero_operand;
                 emitter_sp++;
                 stack_size++;
@@ -1061,7 +1278,7 @@ static void emit_token(token_data t)
 
     op_node_number = ET_used++;
     if (op_node_number == MAX_EXPRESSION_NODES)
-        memoryerror("MAX_EXPRESSION_NODES3", MAX_EXPRESSION_NODES);
+        memoryerror("MAX_EXPRESSION_NODES", MAX_EXPRESSION_NODES);
 
     ET[op_node_number].operator_number = t.value;
     ET[op_node_number].up = -1;
@@ -1082,7 +1299,7 @@ static void emit_token(token_data t)
         else
         {   operand_node_number = ET_used++;
             if (operand_node_number == MAX_EXPRESSION_NODES)
-                memoryerror("MAX_EXPRESSION_NODES4", MAX_EXPRESSION_NODES);
+                memoryerror("MAX_EXPRESSION_NODES", MAX_EXPRESSION_NODES);
             ET[operand_node_number].down = -1;
             ET[operand_node_number].value = emitter_stack[i];
         }
@@ -1102,7 +1319,10 @@ static void emit_token(token_data t)
     emitter_stack[emitter_sp - 1].type = EXPRESSION_OT;
     emitter_stack[emitter_sp - 1].value = op_node_number;
     emitter_stack[emitter_sp - 1].marker = 0;
-    emitter_markers[emitter_sp - 1] = mark_result;
+    emitter_markers[emitter_sp - 1] = 0;
+    emitter_bracket_counts[emitter_sp - 1] = 0;
+    /* Remove the marker for the brackets implied by operator precedence */
+    remove_bracket_layer_from_emitter_stack();
 
     return;
 
@@ -1162,12 +1382,16 @@ the range -32768 to +32767:", folding_error);
     emitter_stack[emitter_sp - 1].value = x;
     emitter_stack[emitter_sp - 1].marker = 0;
     emitter_markers[emitter_sp - 1] = 0;
+    emitter_bracket_counts[emitter_sp - 1] = 0;
 
     if (expr_trace_level >= 2)
     {   printf("Folding constant to: ");
         print_operand(emitter_stack[emitter_sp - 1]);
         printf("\n");
     }
+
+    /* Remove the marker for the brackets implied by operator precedence */
+    remove_bracket_layer_from_emitter_stack();
     return;
 }
 
@@ -1251,7 +1475,7 @@ static void check_property_operator(int from_node)
 static void check_lvalues(int from_node)
 {   int below = ET[from_node].down;
     int opnum = ET[from_node].operator_number, opnum_below;
-    int lvalue_form, i, j;
+    int lvalue_form, i, j = 0;
 
     if (below != -1)
     {
@@ -1411,7 +1635,7 @@ static void insert_exp_to_cond(int n, int context)
     {   if (context==CONDITION_CONTEXT)
         {   new = ET_used++;
             if (new == MAX_EXPRESSION_NODES)
-                memoryerror("MAX_EXPRESSION_NODES5", MAX_EXPRESSION_NODES);
+                memoryerror("MAX_EXPRESSION_NODES", MAX_EXPRESSION_NODES);
             ET[new] = ET[n];
             ET[n].down = new; ET[n].operator_number = NONZERO_OP;
             ET[new].up = n; ET[new].right = -1;
@@ -1434,7 +1658,7 @@ static void insert_exp_to_cond(int n, int context)
 
             new = ET_used++;
             if (new == MAX_EXPRESSION_NODES)
-                memoryerror("MAX_EXPRESSION_NODES6", MAX_EXPRESSION_NODES);
+                memoryerror("MAX_EXPRESSION_NODES", MAX_EXPRESSION_NODES);
             ET[new] = ET[n];
             ET[n].down = new; ET[n].operator_number = NONZERO_OP;
             ET[new].up = n; ET[new].right = -1;
@@ -1493,7 +1717,7 @@ static void func_args_on_stack(int n, int context)
         if (etoken_num_children(pn) > (unsigned int)(opnum == FCALL_OP ? 4:3)) {
           new = ET_used++;
           if (new == MAX_EXPRESSION_NODES)
-            memoryerror("MAX_EXPRESSION_NODES9", MAX_EXPRESSION_NODES);
+            memoryerror("MAX_EXPRESSION_NODES", MAX_EXPRESSION_NODES);
           ET[new] = ET[n];
           ET[n].down = new; 
           ET[n].operator_number = PUSH_OP;
@@ -1516,14 +1740,13 @@ static assembly_operand check_conditions(assembly_operand AO, int context)
     {   if (context != CONDITION_CONTEXT) return AO;
         n = ET_used++;
         if (n == MAX_EXPRESSION_NODES)
-            memoryerror("MAX_EXPRESSION_NODES7", MAX_EXPRESSION_NODES);
+            memoryerror("MAX_EXPRESSION_NODES", MAX_EXPRESSION_NODES);
         ET[n].down = -1;
         ET[n].up = -1;
         ET[n].right = -1;
         ET[n].value = AO;
-        AO.type = EXPRESSION_OT;
+        INITAOT(&AO, EXPRESSION_OT);
         AO.value = n;
-        AO.marker = 0;
     }
 
     insert_exp_to_cond(AO.value, context);
@@ -1652,7 +1875,8 @@ extern assembly_operand parse_expression(int context)
 
         if ((a.type == ENDEXP_TT) && (b.type == ENDEXP_TT))
         {   if (emitter_sp == 0)
-            {   compiler_error("SR error: emitter stack empty");
+            {   error("No expression between brackets '(' and ')'");
+                put_token_back();
                 return AO;
             }
             if (emitter_sp > 1)
@@ -1698,21 +1922,50 @@ extern assembly_operand parse_expression(int context)
             case LOWER_P:
             case EQUAL_P:
                 if (sr_sp == MAX_EXPRESSION_NODES)
-                    memoryerror("MAX_EXPRESSION_NODES8", MAX_EXPRESSION_NODES);
+                    memoryerror("MAX_EXPRESSION_NODES", MAX_EXPRESSION_NODES);
                 sr_stack[sr_sp++] = b;
-                if ((b.type == OP_TT) && (b.value == OR_OP))
+                switch(b.type)
                 {
-                    if (!((sr_stack[sr_sp-2].type == OP_TT) &&
-                          (sr_stack[sr_sp-2].value >= ZERO_OP) &&
-                          (sr_stack[sr_sp-2].value <= NOTPROVIDES_OP)))
-                    {
-                error("'or' not between values to the right of a condition");
-
-                        /* Convert to + for error recovery purposes */
-                        sr_stack[sr_sp-1].value = PLUS_OP;
-                    }
+                    case SUBOPEN_TT:
+                        if (sr_sp >= 2 && sr_stack[sr_sp-2].type == OP_TT && sr_stack[sr_sp-2].value == FCALL_OP)
+                            mark_top_of_emitter_stack(FUNCTION_VALUE_MARKER, b);
+                        else
+                            add_bracket_layer_to_emitter_stack(0);
+                        break;
+                    case OP_TT:
+                        switch(b.value){
+                            case OR_OP:
+                                if (sr_stack[sr_sp-2].type == OP_TT &&
+                                    operators[sr_stack[sr_sp-2].value].precedence == 3)
+                                    mark_top_of_emitter_stack(OR_VALUE_MARKER, b);
+                                else
+                                {   error("'or' not between values to the right of a condition");
+                                    /* Convert to + for error recovery purposes */
+                                    sr_stack[sr_sp-1].value = PLUS_OP;
+                                }
+                                break;
+                            case COMMA_OP:
+                                {
+                                    /* A comma separates arguments only if the shallowest open bracket belongs to a function call. */
+                                    int shallowest_open_bracket_index = sr_sp - 2;
+                                    while (shallowest_open_bracket_index > 0 && sr_stack[shallowest_open_bracket_index].type != SUBOPEN_TT)
+                                        --shallowest_open_bracket_index;
+                                    if (shallowest_open_bracket_index > 0 &&
+                                        sr_stack[shallowest_open_bracket_index-1].type == OP_TT &&
+                                        sr_stack[shallowest_open_bracket_index-1].value == FCALL_OP)
+                                    {   mark_top_of_emitter_stack(ARGUMENT_VALUE_MARKER, b);
+                                        break;
+                                    }
+                                    /* Non-argument-separating commas get treated like any other operator; we fall through to the default case. */
+                                }
+                            default:
+                                {
+                                    /* Add a marker for the brackets implied by operator precedence */
+                                    int operands_on_left = (operators[b.value].usage == PRE_U) ? 0 : 1;
+                                    add_bracket_layer_to_emitter_stack(operands_on_left);
+                                }
+                        }
                 }
-
                 get_next_etoken();
                 break;
             case GREATER_P:
@@ -1792,8 +2045,10 @@ extern void expressp_begin_pass(void)
 extern void expressp_allocate_arrays(void)
 {   ET = my_calloc(sizeof(expression_tree_node), MAX_EXPRESSION_NODES,
         "expression parse trees");
-    emitter_markers = my_calloc(sizeof(char), MAX_EXPRESSION_NODES,
+    emitter_markers = my_calloc(sizeof(int), MAX_EXPRESSION_NODES,
         "emitter markers");
+    emitter_bracket_counts = my_calloc(sizeof(int), MAX_EXPRESSION_NODES,
+        "emitter bracket layer counts");
     emitter_stack = my_calloc(sizeof(assembly_operand), MAX_EXPRESSION_NODES,
         "emitter stack");
     sr_stack = my_calloc(sizeof(token_data), MAX_EXPRESSION_NODES,
@@ -1803,6 +2058,7 @@ extern void expressp_allocate_arrays(void)
 extern void expressp_free_arrays(void)
 {   my_free(&ET, "expression parse trees");
     my_free(&emitter_markers, "emitter markers");
+    my_free(&emitter_bracket_counts, "emitter bracket layer counts");
     my_free(&emitter_stack, "emitter stack");
     my_free(&sr_stack, "shift-reduce parser stack");
 }
