@@ -2,8 +2,8 @@
 /*   "memory" : Memory management and ICL memory setting commands            */
 /*              (For "memoryerror", see "errors.c")                          */
 /*                                                                           */
-/*   Part of Inform 6.30                                                     */
-/*   copyright (c) Graham Nelson 1993 - 2004                                 */
+/*   Part of Inform 6.34                                                     */
+/*   copyright (c) Graham Nelson 1993 - 2017                                 */
 /*                                                                           */
 /* ------------------------------------------------------------------------- */
 
@@ -12,6 +12,7 @@
 int32 malloced_bytes=0;                /* Total amount of memory allocated   */
 
 #ifdef PC_QUICKC
+
 extern void *my_malloc(int32 size, char *whatfor)
 {   char _huge *c;
     if (memout_switch)
@@ -21,6 +22,26 @@ extern void *my_malloc(int32 size, char *whatfor)
     if (c==0) memory_out_error(size, 1, whatfor);
     return(c);
 }
+
+extern void my_realloc(void *pointer, int32 oldsize, int32 size, 
+    char *whatfor)
+{   char _huge *c;
+    if (size==0) {
+        my_free(pointer, whatfor);
+        return;
+    }
+    c=halloc(size,1); malloced_bytes+=size;
+    if (c==0) memory_out_error(size, 1, whatfor);
+    if (memout_switch)
+        printf("Increasing allocation to %ld bytes for %s was (%08lx) \
+now (%08lx)\n",
+            (long int) size,whatfor,(long int) (*(int **)pointer), 
+            (long int) c);
+    memcpy(c, *(int **)pointer, MIN(oldsize, size));
+    hfree(*(int **)pointer);
+    *(int **)pointer = c;
+}
+
 extern void *my_calloc(int32 size, int32 howmany, char *whatfor)
 {   void _huge *c;
     if (memout_switch)
@@ -31,7 +52,29 @@ extern void *my_calloc(int32 size, int32 howmany, char *whatfor)
     if (c==0) memory_out_error(size, howmany, whatfor);
     return(c);
 }
+
+extern void my_recalloc(void *pointer, int32 size, int32 oldhowmany, 
+    int32 howmany, char *whatfor)
+{   void _huge *c;
+    if (size*howmany==0) {
+        my_free(pointer, whatfor);
+        return;
+    }
+    c=(void _huge *)halloc(size*howmany,1); malloced_bytes+=size*howmany;
+    if (c==0) memory_out_error(size, howmany, whatfor);
+    if (memout_switch)
+        printf("Increasing allocation to %ld bytes: array (%ld entries size %ld) \
+for %s was (%08lx) now (%08lx)\n",
+            ((long int)size) * ((long int)howmany),
+            (long int)howmany,(long int)size,whatfor,
+            (long int) *(int **)pointer, (long int) c);
+    memcpy(c, *(int **)pointer, MIN(size*oldhowmany, size*howmany));
+    hfree(*(int **)pointer);
+    *(int **)pointer = c;
+}
+
 #else
+
 extern void *my_malloc(int32 size, char *whatfor)
 {   char *c;
     if (size==0) return(NULL);
@@ -42,6 +85,24 @@ extern void *my_malloc(int32 size, char *whatfor)
             (long int) size,whatfor,(long int) c);
     return(c);
 }
+
+extern void my_realloc(void *pointer, int32 oldsize, int32 size, 
+    char *whatfor)
+{   void *c;
+    if (size==0) {
+        my_free(pointer, whatfor);
+        return;
+    }
+    c=realloc(*(int **)pointer, (size_t) size); malloced_bytes+=size;
+    if (c==0) memory_out_error(size, 1, whatfor);
+    if (memout_switch)
+        printf("Increasing allocation to %ld bytes for %s was (%08lx) \
+now (%08lx)\n",
+            (long int) size,whatfor,(long int) (*(int **)pointer), 
+            (long int) c);
+    *(int **)pointer = c;
+}
+
 extern void *my_calloc(int32 size, int32 howmany, char *whatfor)
 {   void *c;
     if (size*howmany==0) return(NULL);
@@ -55,6 +116,26 @@ for %s at (%08lx)\n",
             (long int) c);
     return(c);
 }
+
+extern void my_recalloc(void *pointer, int32 size, int32 oldhowmany, 
+    int32 howmany, char *whatfor)
+{   void *c;
+    if (size*howmany==0) {
+        my_free(pointer, whatfor);
+        return;
+    }
+    c=realloc(*(int **)pointer, (size_t)size*(size_t)howmany); 
+    malloced_bytes+=size*howmany;
+    if (c==0) memory_out_error(size, howmany, whatfor);
+    if (memout_switch)
+        printf("Increasing allocation to %ld bytes: array (%ld entries size %ld) \
+for %s was (%08lx) now (%08lx)\n",
+            ((long int)size) * ((long int)howmany),
+            (long int)howmany,(long int)size,whatfor,
+            (long int) *(int **)pointer, (long int) c);
+    *(int **)pointer = c;
+}
+
 #endif
 
 extern void my_free(void *pointer, char *whatitwas)
@@ -123,7 +204,7 @@ extern void write_byte_to_memory_block(memory_block *MB, int32 index, int value)
     {   compiler_error_named("memory: negative index to", chunk_name(MB, 0));
         return;
     }
-    if (ch >= 72) fatalerror("One of the memory blocks has exceeded 640K");
+    if (ch >= 72) memoryerror("ALLOC_CHUNK_SIZE", ALLOC_CHUNK_SIZE);
 
     if (MB->chunk[ch] == NULL)
     {   int i;
@@ -162,7 +243,6 @@ int32 MAX_ZCODE_SIZE;
 int MAX_LOW_STRINGS;
 int32 MAX_TRANSCRIPT_SIZE;
 int MAX_CLASSES;
-int MAX_CLASS_TABLE_SIZE;
 int32 MAX_LINK_DATA_SIZE;
 int MAX_INCLUSION_DEPTH;
 int MAX_SOURCE_FILES;
@@ -171,9 +251,20 @@ int32 MAX_OBJ_PROP_TABLE_SIZE;
 int MAX_OBJ_PROP_COUNT;
 int MAX_LOCAL_VARIABLES;
 int MAX_GLOBAL_VARIABLES;
-int DICT_WORD_SIZE;
+int DICT_WORD_SIZE; /* number of characters in a dict word */
+int DICT_CHAR_SIZE; /* (glulx) 1 for one-byte chars, 4 for Unicode chars */
+int DICT_WORD_BYTES; /* DICT_WORD_SIZE*DICT_CHAR_SIZE */
+int ZCODE_HEADER_EXT_WORDS; /* (zcode 1.0) requested header extension size */
+int ZCODE_HEADER_FLAGS_3; /* (zcode 1.1) value to place in Flags 3 word */
 int NUM_ATTR_BYTES;
+int GLULX_OBJECT_EXT_BYTES; /* (glulx) extra bytes for each object record */
 int32 MAX_NUM_STATIC_STRINGS;
+int32 MAX_UNICODE_CHARS;
+int32 MAX_STACK_SIZE;
+int32 MEMORY_MAP_EXTENSION;
+int ALLOC_CHUNK_SIZE;
+int WARN_UNUSED_ROUTINES; /* 0: no, 1: yes except in system files, 2: yes always */
+int OMIT_UNUSED_ROUTINES; /* 0: no, 1: yes */
 
 /* The way memory sizes are set causes great nuisance for those parameters
    which have different defaults under Z-code and Glulx. We have to get
@@ -185,6 +276,7 @@ static int MAX_GLOBAL_VARIABLES_z, MAX_GLOBAL_VARIABLES_g;
 static int MAX_LOCAL_VARIABLES_z, MAX_LOCAL_VARIABLES_g;
 static int DICT_WORD_SIZE_z, DICT_WORD_SIZE_g;
 static int NUM_ATTR_BYTES_z, NUM_ATTR_BYTES_g;
+static int ALLOC_CHUNK_SIZE_z, ALLOC_CHUNK_SIZE_g;
 
 /* ------------------------------------------------------------------------- */
 /*   Memory control from the command line                                    */
@@ -197,38 +289,64 @@ static void list_memory_sizes(void)
     printf("|  %25s = %-7d |\n","MAX_ABBREVS",MAX_ABBREVS);
     printf("|  %25s = %-7d |\n","MAX_ACTIONS",MAX_ACTIONS);
     printf("|  %25s = %-7d |\n","MAX_ADJECTIVES",MAX_ADJECTIVES);
+    printf("|  %25s = %-7d |\n","ALLOC_CHUNK_SIZE",ALLOC_CHUNK_SIZE);
+    printf("|  %25s = %-7d |\n","MAX_ARRAYS",MAX_ARRAYS);
     printf("|  %25s = %-7d |\n","NUM_ATTR_BYTES",NUM_ATTR_BYTES);
     printf("|  %25s = %-7d |\n","MAX_CLASSES",MAX_CLASSES);
-    printf("|  %25s = %-7d |\n","MAX_CLASS_TABLE_SIZE",MAX_CLASS_TABLE_SIZE);
     printf("|  %25s = %-7d |\n","MAX_DICT_ENTRIES",MAX_DICT_ENTRIES);
     printf("|  %25s = %-7d |\n","DICT_WORD_SIZE",DICT_WORD_SIZE);
+    if (glulx_mode)
+      printf("|  %25s = %-7d |\n","DICT_CHAR_SIZE",DICT_CHAR_SIZE);
     printf("|  %25s = %-7d |\n","MAX_EXPRESSION_NODES",MAX_EXPRESSION_NODES);
     printf("|  %25s = %-7d |\n","MAX_GLOBAL_VARIABLES",MAX_GLOBAL_VARIABLES);
     printf("|  %25s = %-7d |\n","HASH_TAB_SIZE",HASH_TAB_SIZE);
+    if (!glulx_mode)
+      printf("|  %25s = %-7d |\n","ZCODE_HEADER_EXT_WORDS",ZCODE_HEADER_EXT_WORDS);
+    if (!glulx_mode)
+      printf("|  %25s = %-7d |\n","ZCODE_HEADER_FLAGS_3",ZCODE_HEADER_FLAGS_3);
     printf("|  %25s = %-7d |\n","MAX_INCLUSION_DEPTH",MAX_INCLUSION_DEPTH);
     printf("|  %25s = %-7d |\n","MAX_INDIV_PROP_TABLE_SIZE",
         MAX_INDIV_PROP_TABLE_SIZE);
     printf("|  %25s = %-7d |\n","MAX_LABELS",MAX_LABELS);
     printf("|  %25s = %-7d |\n","MAX_LINESPACE",MAX_LINESPACE);
     printf("|  %25s = %-7d |\n","MAX_LINK_DATA_SIZE",MAX_LINK_DATA_SIZE);
-    printf("|  %25s = %-7d |\n","MAX_LOCAL_VARIABLES",MAX_LOCAL_VARIABLES);
+    if (glulx_mode)
+      printf("|  %25s = %-7d |\n","MAX_LOCAL_VARIABLES",MAX_LOCAL_VARIABLES);
     printf("|  %25s = %-7d |\n","MAX_LOW_STRINGS",MAX_LOW_STRINGS);
-    printf("|  %25s = %-7d |\n","MAX_NUM_STATIC_STRINGS",MAX_NUM_STATIC_STRINGS);
+    if (glulx_mode)
+      printf("|  %25s = %-7d |\n","MEMORY_MAP_EXTENSION",
+        MEMORY_MAP_EXTENSION);
+    if (glulx_mode)
+      printf("|  %25s = %-7d |\n","MAX_NUM_STATIC_STRINGS",
+        MAX_NUM_STATIC_STRINGS);
     printf("|  %25s = %-7d |\n","MAX_OBJECTS",MAX_OBJECTS);
-    printf("|  %25s = %-7d |\n","MAX_OBJ_PROP_COUNT",
+    if (glulx_mode)
+      printf("|  %25s = %-7d |\n","GLULX_OBJECT_EXT_BYTES",
+        GLULX_OBJECT_EXT_BYTES);
+    if (glulx_mode)
+      printf("|  %25s = %-7d |\n","MAX_OBJ_PROP_COUNT",
         MAX_OBJ_PROP_COUNT);
-    printf("|  %25s = %-7d |\n","MAX_OBJ_PROP_TABLE_SIZE",
+    if (glulx_mode)
+      printf("|  %25s = %-7d |\n","MAX_OBJ_PROP_TABLE_SIZE",
         MAX_OBJ_PROP_TABLE_SIZE);
     printf("|  %25s = %-7d |\n","MAX_PROP_TABLE_SIZE",MAX_PROP_TABLE_SIZE);
     printf("|  %25s = %-7d |\n","MAX_QTEXT_SIZE",MAX_QTEXT_SIZE);
     printf("|  %25s = %-7d |\n","MAX_SOURCE_FILES",MAX_SOURCE_FILES);
-    printf("|  %25s = %-7d |\n","MAX_SYMBOLS",MAX_SYMBOLS);
+    if (glulx_mode)
+      printf("|  %25s = %-7ld |\n","MAX_STACK_SIZE",
+           (long int) MAX_STACK_SIZE);
     printf("|  %25s = %-7d |\n","MAX_STATIC_DATA",MAX_STATIC_DATA);
     printf("|  %25s = %-7ld |\n","MAX_STATIC_STRINGS",
            (long int) MAX_STATIC_STRINGS);
+    printf("|  %25s = %-7d |\n","MAX_SYMBOLS",MAX_SYMBOLS);
     printf("|  %25s = %-7d |\n","SYMBOLS_CHUNK_SIZE",SYMBOLS_CHUNK_SIZE);
     printf("|  %25s = %-7ld |\n","MAX_TRANSCRIPT_SIZE",
            (long int) MAX_TRANSCRIPT_SIZE);
+    if (glulx_mode)
+      printf("|  %25s = %-7ld |\n","MAX_UNICODE_CHARS",
+           (long int) MAX_UNICODE_CHARS);
+    printf("|  %25s = %-7d |\n","WARN_UNUSED_ROUTINES",WARN_UNUSED_ROUTINES);
+    printf("|  %25s = %-7d |\n","OMIT_UNUSED_ROUTINES",OMIT_UNUSED_ROUTINES);
     printf("|  %25s = %-7d |\n","MAX_VERBS",MAX_VERBS);
     printf("|  %25s = %-7d |\n","MAX_VERBSPACE",MAX_VERBSPACE);
     printf("|  %25s = %-7ld |\n","MAX_ZCODE_SIZE",
@@ -275,7 +393,6 @@ extern void set_memory_sizes(int size_flag)
         MAX_NUM_STATIC_STRINGS = 20000;
 
         MAX_CLASSES = 64;
-        MAX_CLASS_TABLE_SIZE = 1000;
 
         MAX_OBJ_PROP_COUNT = 128;
         MAX_OBJ_PROP_TABLE_SIZE = 4096;
@@ -285,6 +402,9 @@ extern void set_memory_sizes(int size_flag)
 
         MAX_GLOBAL_VARIABLES_z = 240;
         MAX_GLOBAL_VARIABLES_g = 512;
+        
+        ALLOC_CHUNK_SIZE_z = 8192;
+        ALLOC_CHUNK_SIZE_g = 32768;
     }
     if (size_flag == LARGE_SIZE)
     {
@@ -323,7 +443,6 @@ extern void set_memory_sizes(int size_flag)
         MAX_NUM_STATIC_STRINGS = 20000;
 
         MAX_CLASSES = 64;
-        MAX_CLASS_TABLE_SIZE = 1000;
 
         MAX_OBJ_PROP_COUNT = 64;
         MAX_OBJ_PROP_TABLE_SIZE = 2048;
@@ -333,6 +452,9 @@ extern void set_memory_sizes(int size_flag)
 
         MAX_GLOBAL_VARIABLES_z = 240;
         MAX_GLOBAL_VARIABLES_g = 512;
+        
+        ALLOC_CHUNK_SIZE_z = 8192;
+        ALLOC_CHUNK_SIZE_g = 16384;
     }
     if (size_flag == SMALL_SIZE)
     {
@@ -371,7 +493,6 @@ extern void set_memory_sizes(int size_flag)
         MAX_NUM_STATIC_STRINGS = 10000;
 
         MAX_CLASSES = 32;
-        MAX_CLASS_TABLE_SIZE = 800;
 
         MAX_OBJ_PROP_COUNT = 64;
         MAX_OBJ_PROP_TABLE_SIZE = 1024;
@@ -381,6 +502,9 @@ extern void set_memory_sizes(int size_flag)
 
         MAX_GLOBAL_VARIABLES_z = 240;
         MAX_GLOBAL_VARIABLES_g = 256;
+        
+        ALLOC_CHUNK_SIZE_z = 8192;
+        ALLOC_CHUNK_SIZE_g = 8192;
     }
 
     /* Regardless of size_flag... */
@@ -388,10 +512,27 @@ extern void set_memory_sizes(int size_flag)
     MAX_INCLUSION_DEPTH = 5;
     MAX_LOCAL_VARIABLES_z = 16;
     MAX_LOCAL_VARIABLES_g = 32;
+    DICT_CHAR_SIZE = 1;
     DICT_WORD_SIZE_z = 6;
     DICT_WORD_SIZE_g = 9;
     NUM_ATTR_BYTES_z = 6;
     NUM_ATTR_BYTES_g = 7;
+    /* Backwards-compatible behavior: allow for a unicode table
+       whether we need one or not. The user can set this to zero if
+       there's no unicode table. */
+    ZCODE_HEADER_EXT_WORDS = 3;
+    ZCODE_HEADER_FLAGS_3 = 0;
+    GLULX_OBJECT_EXT_BYTES = 0;
+    MAX_UNICODE_CHARS = 64;
+    MEMORY_MAP_EXTENSION = 0;
+    /* We estimate the default Glulx stack size at 4096. That's about
+       enough for 90 nested function calls with 8 locals each -- the
+       same capacity as the Z-Spec's suggestion for Z-machine stack
+       size. Note that Inform 7 wants more stack; I7-generated code
+       sets MAX_STACK_SIZE to 65536 by default. */
+    MAX_STACK_SIZE = 4096;
+    OMIT_UNUSED_ROUTINES = 0;
+    WARN_UNUSED_ROUTINES = 0;
 
     adjust_memory_sizes();
 }
@@ -405,6 +546,7 @@ extern void adjust_memory_sizes()
     MAX_LOCAL_VARIABLES = MAX_LOCAL_VARIABLES_z;
     DICT_WORD_SIZE = DICT_WORD_SIZE_z;
     NUM_ATTR_BYTES = NUM_ATTR_BYTES_z;
+    ALLOC_CHUNK_SIZE = ALLOC_CHUNK_SIZE_z;
   }
   else {
     MAX_ZCODE_SIZE = MAX_ZCODE_SIZE_g;
@@ -413,6 +555,7 @@ extern void adjust_memory_sizes()
     MAX_LOCAL_VARIABLES = MAX_LOCAL_VARIABLES_g;
     DICT_WORD_SIZE = DICT_WORD_SIZE_g;
     NUM_ATTR_BYTES = NUM_ATTR_BYTES_g;
+    ALLOC_CHUNK_SIZE = ALLOC_CHUNK_SIZE_g;
   }
 }
 
@@ -475,14 +618,44 @@ static void explain_parameter(char *command)
   can be any number.\n");
         return;
     }
+    if (strcmp(command,"DICT_CHAR_SIZE")==0)
+    {   printf(
+"  DICT_CHAR_SIZE is the byte size of one character in the dictionary. \n\
+  (This is only meaningful in Glulx, since Z-code has compressed dictionary \n\
+  words.) It can be either 1 (the default) or 4 (to enable full Unicode \n\
+  input.)\n");
+        return;
+    }
     if (strcmp(command,"NUM_ATTR_BYTES")==0)
     {   printf(
 "  NUM_ATTR_BYTES is the space used to store attribute flags. Each byte \n\
-  stores eight attribytes. In Z-code this is always 6 (only 4 are used in \n\
+  stores eight attributes. In Z-code this is always 6 (only 4 are used in \n\
   v3 games). In Glulx it can be any number which is a multiple of four, \n\
   plus three.\n");
         return;
     }
+    if (strcmp(command,"ZCODE_HEADER_EXT_WORDS")==0)
+    {   printf(
+"  ZCODE_HEADER_EXT_WORDS is the number of words in the Z-code header \n\
+  extension table (Z-Spec 1.0). The -W switch also sets this. It defaults \n\
+  to 3, but can be set higher. (It can be set lower if no Unicode \n\
+  translation table is created.)\n");
+        return;
+    }
+    if (strcmp(command,"ZCODE_HEADER_FLAGS_3")==0)
+    {   printf(
+"  ZCODE_HEADER_FLAGS_3 is the value to store in the Flags 3 word of the \n\
+  header extension table (Z-Spec 1.1).\n");
+        return;
+    }
+    if (strcmp(command,"GLULX_OBJECT_EXT_BYTES")==0)
+    {   printf(
+"  GLULX_OBJECT_EXT_BYTES is an amount of additional space to add to each \n\
+  object record. It is initialized to zero bytes, and the game is free to \n\
+  use it as desired. (This is only meaningful in Glulx, since Z-code \n\
+  specifies the object structure.)\n");
+        return;
+    }
     if (strcmp(command,"MAX_STATIC_DATA")==0)
     {   printf(
 "  MAX_STATIC_DATA is the size of an array of integers holding initial \n\
@@ -552,7 +725,6 @@ static void explain_parameter(char *command)
   Inform automatically ensures that this is at least twice the size of\n\
   MAX_QTEXT_SIZE, to be on the safe side.");
         return;
-
     }
     if (strcmp(command,"MAX_ZCODE_SIZE")==0)
     {
@@ -591,12 +763,6 @@ static void explain_parameter(char *command)
   is cheap to increase.\n");
         return;
     }
-    if (strcmp(command,"MAX_CLASS_TABLE_SIZE")==0)
-    {   printf(
-"  MAX_CLASS_TABLE_SIZE is the number of bytes allocated to hold the table \n\
-  of properties to inherit from each class.\n");
-        return;
-    }
     if (strcmp(command,"MAX_INCLUSION_DEPTH")==0)
     {   printf(
 "  MAX_INCLUSION_DEPTH is the number of nested includes permitted.\n");
@@ -644,7 +810,58 @@ static void explain_parameter(char *command)
 "  MAX_NUM_STATIC_STRINGS is the maximum number of compiled strings \n\
   allowed in the program. (Glulx only)\n");
         return;
-
+    }
+    if (strcmp(command,"MAX_UNICODE_CHARS")==0)
+    {
+        printf(
+"  MAX_UNICODE_CHARS is the maximum number of different Unicode characters \n\
+  (beyond the Latin-1 range, $00..$FF) which the game text can use. \n\
+  (Glulx only)\n");
+        return;
+    }
+    if (strcmp(command,"ALLOC_CHUNK_SIZE")==0)
+    {
+        printf(
+"  ALLOC_CHUNK_SIZE is a base unit of Inform's internal memory allocation \n\
+  for various structures.\n");
+        return;
+    }
+    if (strcmp(command,"MAX_STACK_SIZE")==0)
+    {
+        printf(
+"  MAX_STACK_SIZE is the maximum size (in bytes) of the interpreter stack \n\
+  during gameplay. (Glulx only)\n");
+        return;
+    }
+    if (strcmp(command,"MEMORY_MAP_EXTENSION")==0)
+    {
+        printf(
+"  MEMORY_MAP_EXTENSION is the number of bytes (all zeroes) to map into \n\
+  memory after the game file. (Glulx only)\n");
+        return;
+    }
+    if (strcmp(command,"WARN_UNUSED_ROUTINES")==0)
+    {
+        printf(
+"  WARN_UNUSED_ROUTINES, if set to 2, will display a warning for each \n\
+  routine in the game file which is never called. (This includes \n\
+  routines called only from uncalled routines, etc.) If set to 1, will warn \n\
+  only about functions in game code, not in the system library.\n");
+        return;
+    }
+    if (strcmp(command,"OMIT_UNUSED_ROUTINES")==0)
+    {
+        printf(
+"  OMIT_UNUSED_ROUTINES, if set to 1, will avoid compiling unused routines \n\
+  into the game file.\n");
+        return;
+    }
+    if (strcmp(command,"SERIAL")==0)
+    {
+        printf(
+"  SERIAL, if set, will be used as the six digit serial number written into \n\
+  the header of the output file.\n");
+        return;
     }
 
     printf("No such memory setting as \"%s\"\n",command);
@@ -652,6 +869,59 @@ static void explain_parameter(char *command)
     return;
 }
 
+/* Parse a decimal number as an int32. Return true if a valid number
+   was found; otherwise print a warning and return false.
+
+   Anything over nine digits is considered an overflow; we report a
+   warning but return +/- 999999999 (and true). This is not entirely
+   clever about leading zeroes ("0000000001" is treated as an
+   overflow) but this is better than trying to detect genuine
+   overflows in a long.
+
+   (Some Glulx settings might conceivably want to go up to $7FFFFFFF,
+   which is a ten-digit number, but we're not going to allow that
+   today.)
+
+   This used to rely on atoi(), and we retain the atoi() behavior of
+   ignoring garbage characters after a valid decimal number.
+ */
+static int parse_memory_setting(char *str, char *label, int32 *result)
+{
+    char *cx = str;
+    char *ex;
+    long val;
+
+    *result = 0;
+
+    while (*cx == ' ') cx++;
+
+    val = strtol(cx, &ex, 10);    
+
+    if (ex == cx) {
+        printf("Bad numerical setting in $ command \"%s=%s\"\n",
+            label, str);
+        return 0;
+    }
+
+    if (*cx == '-') {
+        if (ex > cx+10) {
+            val = -999999999;
+            printf("Numerical setting underflowed in $ command \"%s=%s\" (limiting to %ld)\n",
+                label, str, val);
+        }
+    }
+    else {
+        if (ex > cx+9) {
+            val = 999999999;
+            printf("Numerical setting overflowed in $ command \"%s=%s\" (limiting to %ld)\n",
+                label, str, val);
+        }
+    }
+
+    *result = (int32)val;
+    return 1;
+}
+
 extern void memory_command(char *command)
 {   int i, k, flag=0; int32 j;
 
@@ -667,10 +937,7 @@ extern void memory_command(char *command)
     for (i=0; command[i]!=0; i++)
     {   if (command[i]=='=')
         {   command[i]=0;
-            j=(int32) atoi(command+i+1);
-            if ((j==0) && (command[i+1]!='0'))
-            {   printf("Bad numerical setting in $ command \"%s=%s\"\n",
-                    command,command+i+1);
+            if (!parse_memory_setting(command+i+1, command, &j)) {
                 return;
             }
             if (strcmp(command,"BUFFER_LENGTH")==0)
@@ -702,10 +969,18 @@ extern void memory_command(char *command)
             {   DICT_WORD_SIZE=j, flag=1;
                 DICT_WORD_SIZE_g=DICT_WORD_SIZE_z=j;
             }
+            if (strcmp(command,"DICT_CHAR_SIZE")==0)
+                DICT_CHAR_SIZE=j, flag=1;
             if (strcmp(command,"NUM_ATTR_BYTES")==0) 
             {   NUM_ATTR_BYTES=j, flag=1;
                 NUM_ATTR_BYTES_g=NUM_ATTR_BYTES_z=j;
             }
+            if (strcmp(command,"ZCODE_HEADER_EXT_WORDS")==0)
+                ZCODE_HEADER_EXT_WORDS=j, flag=1;
+            if (strcmp(command,"ZCODE_HEADER_FLAGS_3")==0)
+                ZCODE_HEADER_FLAGS_3=j, flag=1;
+            if (strcmp(command,"GLULX_OBJECT_EXT_BYTES")==0)
+                GLULX_OBJECT_EXT_BYTES=j, flag=1;
             if (strcmp(command,"MAX_STATIC_DATA")==0)
                 MAX_STATIC_DATA=j, flag=1;
             if (strcmp(command,"MAX_OLDEPTH")==0)
@@ -759,8 +1034,6 @@ extern void memory_command(char *command)
                 MAX_TRANSCRIPT_SIZE=j, flag=1;
             if (strcmp(command,"MAX_CLASSES")==0)
                 MAX_CLASSES=j, flag=1;
-            if (strcmp(command,"MAX_CLASS_TABLE_SIZE")==0)
-                MAX_CLASS_TABLE_SIZE=j, flag=1;
             if (strcmp(command,"MAX_INCLUSION_DEPTH")==0)
                 MAX_INCLUSION_DEPTH=j, flag=1;
             if (strcmp(command,"MAX_SOURCE_FILES")==0)
@@ -779,6 +1052,45 @@ extern void memory_command(char *command)
             {   MAX_GLOBAL_VARIABLES=j, flag=1;
                 MAX_GLOBAL_VARIABLES_g=MAX_GLOBAL_VARIABLES_z=j;
             }
+            if (strcmp(command,"ALLOC_CHUNK_SIZE")==0)
+            {   ALLOC_CHUNK_SIZE=j, flag=1;
+                ALLOC_CHUNK_SIZE_g=ALLOC_CHUNK_SIZE_z=j;
+            }
+            if (strcmp(command,"MAX_UNICODE_CHARS")==0)
+                MAX_UNICODE_CHARS=j, flag=1;
+            if (strcmp(command,"MAX_STACK_SIZE")==0)
+            {
+                MAX_STACK_SIZE=j, flag=1;
+                /* Adjust up to a 256-byte boundary. */
+                MAX_STACK_SIZE = (MAX_STACK_SIZE + 0xFF) & (~0xFF);
+            }
+            if (strcmp(command,"MEMORY_MAP_EXTENSION")==0)
+            {
+                MEMORY_MAP_EXTENSION=j, flag=1;
+                /* Adjust up to a 256-byte boundary. */
+                MEMORY_MAP_EXTENSION = (MEMORY_MAP_EXTENSION + 0xFF) & (~0xFF);
+            }
+            if (strcmp(command,"WARN_UNUSED_ROUTINES")==0)
+            {
+                WARN_UNUSED_ROUTINES=j, flag=1;
+                if (WARN_UNUSED_ROUTINES > 2 || WARN_UNUSED_ROUTINES < 0)
+                    WARN_UNUSED_ROUTINES = 2;
+            }
+            if (strcmp(command,"OMIT_UNUSED_ROUTINES")==0)
+            {
+                OMIT_UNUSED_ROUTINES=j, flag=1;
+                if (OMIT_UNUSED_ROUTINES > 1 || OMIT_UNUSED_ROUTINES < 0)
+                    OMIT_UNUSED_ROUTINES = 1;
+            }
+            if (strcmp(command,"SERIAL")==0)
+            {
+                if (j >= 0 && j <= 999999)
+                {
+                    sprintf(serial_code_buffer,"%06d",j);
+                    serial_code_given_in_program = TRUE;
+                    flag=1;
+                }
+            }
 
             if (flag==0)
                 printf("No such memory setting as \"%s\"\n", command);
