@@ -1,8 +1,8 @@
 /* ------------------------------------------------------------------------- */
 /*   "syntax" : Syntax analyser and compiler                                 */
 /*                                                                           */
-/*   Part of Inform 6.30                                                     */
-/*   copyright (c) Graham Nelson 1993 - 2004                                 */
+/*   Part of Inform 6.34                                                     */
+/*   copyright (c) Graham Nelson 1993 - 2017                                 */
 /*                                                                           */
 /* ------------------------------------------------------------------------- */
 
@@ -46,19 +46,91 @@ static void begin_syntax_line(int statement_mode)
 
     if (debugfile_switch)
     {   get_next_token();
-        debug_line_ref = token_line_ref;
+        statement_debug_location = get_token_location();
         put_token_back();
     }
 }
 
 extern void panic_mode_error_recovery(void)
 {
+    /* Consume tokens until the next semicolon (or end of file).
+       This is typically called after a syntax error, in hopes of
+       getting parsing back on track. */
+
     while ((token_type != EOF_TT)
            && ((token_type != SEP_TT)||(token_value != SEMICOLON_SEP)))
 
         get_next_token();
 }
 
+extern void get_next_token_with_directives(void)
+{
+    /* A higher-level version of get_next_token(), which detects and
+       obeys directives such as #ifdef/#ifnot/#endif. (The # sign is
+       required in this case.)
+
+       This is called while parsing a long construct, such as Class or
+       Object, where we want to support internal #ifdefs. (Although
+       function-parsing predates this and doesn't make use of it.)
+
+       (Technically this permits *any* #-directive, which means you
+       can define global variables or properties or what-have-you in
+       the middle of an object. You can do that in the middle of an
+       object, too. Don't. It's about as well-supported as Wile E.
+       Coyote one beat before the plummet-lines kick in.) */
+
+    int directives_save, segment_markers_save, statements_save;
+
+    while (TRUE)
+    {
+        get_next_token();
+
+        /* If the first token is not a '#', return it directly. */
+        if ((token_type != SEP_TT) || (token_value != HASH_SEP))
+            return;
+
+        /* Save the lexer flags, and set up for directive parsing. */
+        directives_save = directives.enabled;
+        segment_markers_save = segment_markers.enabled;
+        statements_save = statements.enabled;
+
+        directives.enabled = TRUE;
+        segment_markers.enabled = FALSE;
+        statements.enabled = FALSE;
+        conditions.enabled = FALSE;
+        local_variables.enabled = FALSE;
+        misc_keywords.enabled = FALSE;
+        system_functions.enabled = FALSE;
+
+        get_next_token();
+
+        if ((token_type == SEP_TT) && (token_value == OPEN_SQUARE_SEP))
+        {   error("It is illegal to nest a routine inside an object using '#['");
+            return;
+        }
+
+        if (token_type == DIRECTIVE_TT)
+            parse_given_directive(TRUE);
+        else
+        {   ebf_error("directive", token_text);
+            return;
+        }
+
+        /* Restore all the lexer flags. (We are squashing several of them
+           into a single save variable, which I think is safe because that's
+           what CKnight did.)
+        */
+        directive_keywords.enabled = FALSE;
+        directives.enabled = directives_save;
+        segment_markers.enabled = segment_markers_save;
+        statements.enabled =
+            conditions.enabled =
+            local_variables.enabled =
+            misc_keywords.enabled = 
+            system_functions.enabled = statements_save;
+    }
+}
+
 extern void parse_program(char *source)
 {
     lexical_source = source;
@@ -68,11 +140,13 @@ extern void parse_program(char *source)
 extern int parse_directive(int internal_flag)
 {
     /*  Internal_flag is FALSE if the directive is encountered normally,
-        TRUE if encountered with a # prefix inside a routine.
+        TRUE if encountered with a # prefix inside a routine or object
+        definition.
 
         Returns: TRUE if program continues, FALSE if end of file reached.    */
 
-    int routine_symbol;
+    int routine_symbol, rep_symbol;
+    int is_renamed;
 
     begin_syntax_line(FALSE);
     get_next_token();
@@ -92,7 +166,11 @@ extern int parse_directive(int internal_flag)
         directive_keywords.enabled = FALSE;
         segment_markers.enabled = FALSE;
 
+        /* The upcoming symbol is a definition; don't count it as a
+           top-level reference *to* the function. */
+        df_dont_note_global_symbols = TRUE;
         get_next_token();
+        df_dont_note_global_symbols = FALSE;
         if ((token_type != SYMBOL_TT)
             || ((!(sflags[token_value] & UNKNOWN_SFLAG))
                 && (!(sflags[token_value] & REPLACE_SFLAG))))
@@ -102,8 +180,16 @@ extern int parse_directive(int internal_flag)
 
         routine_symbol = token_value;
 
-        if ((sflags[routine_symbol] & REPLACE_SFLAG) && (is_systemfile()))
-        {   dont_enter_into_symbol_table = TRUE;
+        rep_symbol = routine_symbol;
+        is_renamed = find_symbol_replacement(&rep_symbol);
+
+        if ((sflags[routine_symbol] & REPLACE_SFLAG) 
+            && !is_renamed && (is_systemfile()))
+        {   /* The function is definitely being replaced (system_file
+               always loses priority in a replacement) but is not
+               being renamed to something else. Skip its definition
+               entirely. */
+            dont_enter_into_symbol_table = TRUE;
             do
             {   get_next_token();
             } while (!((token_type == EOF_TT)
@@ -113,12 +199,24 @@ extern int parse_directive(int internal_flag)
             if (token_type == EOF_TT) return FALSE;
         }
         else
-        {   assign_symbol(routine_symbol,
+        {   /* Parse the function definition and assign its symbol. */
+            assign_symbol(routine_symbol,
                 parse_routine(lexical_source, FALSE,
                     (char *) symbs[routine_symbol], FALSE, routine_symbol),
                 ROUTINE_T);
             slines[routine_symbol] = routine_starts_line;
         }
+
+        if (is_renamed) {
+            /* This function was subject to a "Replace X Y" directive.
+               The first time we see a definition for symbol X, we
+               copy it to Y -- that's the "original" form of the
+               function. */
+            if (svals[rep_symbol] == 0) {
+                assign_symbol(rep_symbol, svals[routine_symbol], ROUTINE_T);
+            }
+        }
+
         get_next_token();
         if ((token_type != SEP_TT) || (token_value != SEMICOLON_SEP))
         {   ebf_error("';' after ']'", token_text);
@@ -128,18 +226,27 @@ extern int parse_directive(int internal_flag)
     }
 
     if ((token_type == SYMBOL_TT) && (stypes[token_value] == CLASS_T))
-    {   sflags[token_value] |= USED_SFLAG;
+    {   if (internal_flag)
+        {   error("It is illegal to nest an object in a routine using '#classname'");
+            return(TRUE);
+        }
+        sflags[token_value] |= USED_SFLAG;
         make_object(FALSE, NULL, -1, -1, svals[token_value]);
         return TRUE;
     }
 
     if (token_type != DIRECTIVE_TT)
-    {   ebf_error("directive, '[' or class name", token_text);
+    {   /* If we're internal, we expect only a directive here. If
+           we're top-level, the possibilities are broader. */
+        if (internal_flag)
+            ebf_error("directive", token_text);
+        else
+            ebf_error("directive, '[' or class name", token_text);
         panic_mode_error_recovery();
         return TRUE;
     }
 
-    return !(parse_given_directive());
+    return !(parse_given_directive(internal_flag));
 }
 
 static int switch_sign(void)
@@ -213,10 +320,10 @@ static void parse_switch_spec(assembly_operand switch_value, int label,
 
         if (action_switch)
         {   get_next_token();
-            spec_stack[spec_sp].type = 
-                ((!glulx_mode) ? LONG_CONSTANT_OT : CONSTANT_OT);
-            spec_stack[spec_sp].value = 0;
-            spec_stack[spec_sp].marker = 0;
+            if (token_type == SQ_TT || token_type == DQ_TT) {
+                ebf_error("action (or fake action) name", token_text);
+                continue;
+            }
             spec_stack[spec_sp] = action_of_name(token_text);
 
             if (spec_stack[spec_sp].value == -1)
@@ -312,7 +419,8 @@ extern int32 parse_routine(char *source, int embedded_flag, char *name,
 {   int32 packed_address; int i; int debug_flag = FALSE;
     int switch_clause_made = FALSE, default_clause_made = FALSE,
         switch_label = 0;
-    dbgl start_line_ref = token_line_ref;
+    debug_location_beginning beginning_debug_location =
+        get_token_location_beginning();
 
     /*  (switch_label needs no initialisation here, but it prevents some
         compilers from issuing warnings)   */
@@ -344,6 +452,12 @@ extern int32 parse_routine(char *source, int embedded_flag, char *name,
             break;
         }
 
+        if (strlen(token_text) > MAX_IDENTIFIER_LENGTH)
+        {   error_named("Local variable identifier too long:", token_text);
+            panic_mode_error_recovery();
+            break;
+        }
+
         if (no_locals == MAX_LOCAL_VARIABLES-1)
         {   error_numbered("Too many local variables for a routine; max is",
                 MAX_LOCAL_VARIABLES-1);
@@ -367,7 +481,7 @@ extern int32 parse_routine(char *source, int embedded_flag, char *name,
         sflags[r_symbol] |= STAR_SFLAG;
 
     packed_address = assemble_routine_header(no_locals, debug_flag,
-        name, &start_line_ref, embedded_flag, r_symbol);
+        name, embedded_flag, r_symbol);
 
     do
     {   begin_syntax_line(TRUE);
@@ -376,8 +490,10 @@ extern int32 parse_routine(char *source, int embedded_flag, char *name,
 
         if (token_type == EOF_TT)
         {   ebf_error("']'", token_text);
+            assemble_routine_end
+                (embedded_flag,
+                 get_token_location_end(beginning_debug_location));
             put_token_back();
-            assemble_routine_end(embedded_flag, &token_line_ref);
             break;
         }
 
@@ -387,7 +503,11 @@ extern int32 parse_routine(char *source, int embedded_flag, char *name,
                 assemble_label_no(switch_label);
             directives.enabled = TRUE;
             sequence_point_follows = TRUE;
-            assemble_routine_end(embedded_flag, &token_line_ref);
+            get_next_token();
+            assemble_routine_end
+                (embedded_flag,
+                 get_token_location_end(beginning_debug_location));
+            put_token_back();
             break;
         }
 
@@ -446,12 +566,10 @@ extern int32 parse_routine(char *source, int embedded_flag, char *name,
                 put_token_back(); put_token_back();
 
                 if (!glulx_mode) {
-                    AO.type = VARIABLE_OT; AO.value = 249; AO.marker = 0;
+                    INITAOTV(&AO, VARIABLE_OT, 249);
                 }
                 else {
-                    AO.type = GLOBALVAR_OT;
-                    AO.value = MAX_LOCAL_VARIABLES+6; /* sw__var */
-                    AO.marker = 0;
+                    INITAOTV(&AO, GLOBALVAR_OT, MAX_LOCAL_VARIABLES+6); /* sw__var */
                 }
                 parse_switch_spec(AO, switch_label, TRUE);
 
@@ -472,7 +590,7 @@ extern int32 parse_routine(char *source, int embedded_flag, char *name,
 
 extern void parse_code_block(int break_label, int continue_label,
     int switch_rule)
-{   int switch_clause_made = FALSE, default_clause_made = FALSE, switch_label,
+{   int switch_clause_made = FALSE, default_clause_made = FALSE, switch_label = 0,
         unary_minus_flag;
 
     begin_syntax_line(TRUE);
