@@ -2,8 +2,8 @@
 /*   "verbs" :  Manages actions and grammar tables; parses the directives    */
 /*              Verb and Extend.                                             */
 /*                                                                           */
-/*   Part of Inform 6.30                                                     */
-/*   copyright (c) Graham Nelson 1993 - 2004                                 */
+/*   Part of Inform 6.34                                                     */
+/*   copyright (c) Graham Nelson 1993 - 2017                                 */
 /*                                                                           */
 /* ------------------------------------------------------------------------- */
 
@@ -123,18 +123,22 @@ static void new_action(char *b, int c)
 extern void make_fake_action(void)
 {   int i;
     char action_sub[MAX_IDENTIFIER_LENGTH+4];
+    debug_location_beginning beginning_debug_location =
+        get_token_location_beginning();
 
     get_next_token();
     if (token_type != SYMBOL_TT)
-    {   ebf_error("new fake action name", token_text);
+    {   discard_token_location(beginning_debug_location);
+        ebf_error("new fake action name", token_text);
         panic_mode_error_recovery(); return;
     }
 
-    sprintf(action_sub, "%s__A", token_text);
+    snprintf(action_sub, MAX_IDENTIFIER_LENGTH+4, "%s__A", token_text);
     i = symbol_index(action_sub, -1);
 
     if (!(sflags[i] & UNKNOWN_SFLAG))
-    {   ebf_error("new fake action name", token_text);
+    {   discard_token_location(beginning_debug_location);
+        ebf_error("new fake action name", token_text);
         panic_mode_error_recovery(); return;
     }
 
@@ -142,12 +146,18 @@ extern void make_fake_action(void)
         FAKE_ACTION_T);
 
     new_action(token_text, i);
+
     if (debugfile_switch)
-    {   write_debug_byte(FAKE_ACTION_DBR);
-        write_debug_byte(svals[i]/256);
-        write_debug_byte(svals[i]%256);
-        write_debug_string(token_text);
+    {   debug_file_printf("<fake-action>");
+        debug_file_printf("<identifier>##%s</identifier>", token_text);
+        debug_file_printf("<value>%d</value>", svals[i]);
+        get_next_token();
+        write_debug_locations
+            (get_token_location_end(beginning_debug_location));
+        put_token_back();
+        debug_file_printf("</fake-action>");
     }
+
     return;
 }
 
@@ -160,12 +170,12 @@ extern assembly_operand action_of_name(char *name)
     int j;
     assembly_operand AO;
 
-    sprintf(action_sub, "%s__A", name);
+    snprintf(action_sub, MAX_IDENTIFIER_LENGTH+4, "%s__A", name);
     j = symbol_index(action_sub, -1);
 
     if (stypes[j] == FAKE_ACTION_T)
-    {   AO.value = svals[j];
-        AO.marker = 0;
+    {   INITAO(&AO);
+        AO.value = svals[j];
         if (!glulx_mode)
           AO.type = LONG_CONSTANT_OT;
         else
@@ -184,6 +194,7 @@ extern assembly_operand action_of_name(char *name)
     }
     sflags[j] |= USED_SFLAG;
 
+    INITAO(&AO);
     AO.value = svals[j];
     AO.marker = ACTION_MV;
     if (!glulx_mode) {
@@ -198,20 +209,28 @@ extern assembly_operand action_of_name(char *name)
 
 extern void find_the_actions(void)
 {   int i; int32 j;
+    char action_name[MAX_IDENTIFIER_LENGTH+4];
     char action_sub[MAX_IDENTIFIER_LENGTH+4];
 
     if (module_switch)
         for (i=0; i<no_actions; i++) action_byte_offset[i] = 0;
     else
     for (i=0; i<no_actions; i++)
-    {   strcpy(action_sub, (char *) symbs[action_symbol[i]]);
-        strcpy(action_sub + strlen(action_sub) - 3, "Sub");
+    {   strcpy(action_name, (char *) symbs[action_symbol[i]]);
+        action_name[strlen(action_name) - 3] = '\0'; /* remove "__A" */
+        strcpy(action_sub, action_name);
+        strcat(action_sub, "Sub");
         j = symbol_index(action_sub, -1);
         if (sflags[j] & UNKNOWN_SFLAG)
-            error_named("There is no action routine called", action_sub);
+        {
+            error_named_at("No ...Sub action routine found for action:", action_name, slines[action_symbol[i]]);
+        }
         else
         if (stypes[j] != ROUTINE_T)
-            error_named("Not an action routine:", action_sub);
+        {
+            error_named_at("No ...Sub action routine found for action:", action_name, slines[action_symbol[i]]);
+            error_named_at("-- ...Sub symbol found, but not a routine:", action_sub, slines[j]);
+        }
         else
         {   action_byte_offset[i] = svals[j];
             sflags[j] |= USED_SFLAG;
@@ -235,7 +254,7 @@ static int make_adjective(char *English_word)
         table is left empty in GV2.                                          */
 
     int i; 
-    uchar new_sort_code[MAX_DICT_WORD_SIZE];
+    uchar new_sort_code[MAX_DICT_WORD_BYTES];
 
     if (no_adjectives >= MAX_ADJECTIVES)
         memoryerror("MAX_ADJECTIVES", MAX_ADJECTIVES);
@@ -243,11 +262,11 @@ static int make_adjective(char *English_word)
     dictionary_prepare(English_word, new_sort_code);
     for (i=0; i<no_adjectives; i++)
         if (compare_sorts(new_sort_code,
-          adjective_sort_code+i*DICT_WORD_SIZE) == 0)
+          adjective_sort_code+i*DICT_WORD_BYTES) == 0)
             return(0xff-i);
     adjectives[no_adjectives]
         = dictionary_add(English_word,8,0,0xff-no_adjectives);
-    copy_sorts(adjective_sort_code+no_adjectives*DICT_WORD_SIZE,
+    copy_sorts(adjective_sort_code+no_adjectives*DICT_WORD_BYTES,
         new_sort_code);
     return(0xff-no_adjectives++);
 }
@@ -289,12 +308,13 @@ static int find_or_renumber_verb(char *English_verb, int *new_number)
     char *p;
     p=English_verb_list;
     while (p < English_verb_list_top)
-    {   if (strcmp(English_verb, p+2) == 0)
+    {   if (strcmp(English_verb, p+3) == 0)
         {   if (new_number)
-            {   p[1] = *new_number;
+            {   p[1] = (*new_number)/256;
+                p[2] = (*new_number)%256;
                 return 0;
             }
-            return((uchar)p[1]);
+            return(256*((uchar)p[1]))+((uchar)p[2]);
         }
         p=p+(uchar)p[0];
     }
@@ -305,20 +325,29 @@ static void register_verb(char *English_verb, int number)
 {
     /*  Registers a new English verb as referring to the given Inform-verb
         number.  (See comments above for format of the list.)                */
+    int entrysize;
 
     if (find_or_renumber_verb(English_verb, NULL) != -1)
     {   error_named("Two different verb definitions refer to", English_verb);
         return;
     }
 
-    English_verb_list_size += strlen(English_verb)+3;
+    /* We set a hard limit of MAX_VERB_WORD_SIZE=120 because the
+       English_verb_list table stores length in a leading byte. (We could
+       raise that to 250, really, but there's little point when
+       MAX_DICT_WORD_SIZE is 40.) */
+    entrysize = strlen(English_verb)+4;
+    if (entrysize > MAX_VERB_WORD_SIZE+4)
+        error_numbered("Verb word is too long -- max length is", MAX_VERB_WORD_SIZE);
+    English_verb_list_size += entrysize;
     if (English_verb_list_size >= MAX_VERBSPACE)
         memoryerror("MAX_VERBSPACE", MAX_VERBSPACE);
 
-    English_verb_list_top[0] = 3+strlen(English_verb);
-    English_verb_list_top[1] = number;
-    strcpy(English_verb_list_top+2, English_verb);
-    English_verb_list_top += English_verb_list_top[0];
+    English_verb_list_top[0] = entrysize;
+    English_verb_list_top[1] = number/256;
+    English_verb_list_top[2] = number%256;
+    strcpy(English_verb_list_top+3, English_verb);
+    English_verb_list_top += entrysize;
 }
 
 static int get_verb(void)
@@ -371,11 +400,17 @@ static int grammar_line(int verbnum, int line)
     int j, bytecode, mark; int32 wordcode;
     int grammar_token, slash_mode, last_was_slash;
     int reverse_action, TOKEN_SIZE;
+    debug_location_beginning beginning_debug_location =
+        get_token_location_beginning();
 
     get_next_token();
-    if ((token_type == SEP_TT) && (token_value == SEMICOLON_SEP)) return FALSE;
+    if ((token_type == SEP_TT) && (token_value == SEMICOLON_SEP))
+    {   discard_token_location(beginning_debug_location);
+        return FALSE;
+    }
     if (!((token_type == SEP_TT) && (token_value == TIMES_SEP)))
-    {   ebf_error("'*' divider", token_text);
+    {   discard_token_location(beginning_debug_location);
+        ebf_error("'*' divider", token_text);
         panic_mode_error_recovery();
         return FALSE;
     }
@@ -383,7 +418,8 @@ static int grammar_line(int verbnum, int line)
     /*  Have we run out of lines or token space?  */
 
     if (line >= MAX_LINES_PER_VERB)
-    {   error("Too many lines of grammar for verb. This maximum is built \
+    {   discard_token_location(beginning_debug_location);
+        error("Too many lines of grammar for verb. This maximum is built \
 into Inform, so suggest rewriting grammar using general parsing routines");
         return(FALSE);
     }
@@ -394,11 +430,15 @@ into Inform, so suggest rewriting grammar using general parsing routines");
     mark = grammar_lines_top;
     if (!glulx_mode) {
         if (mark + 100 >= MAX_LINESPACE)
+        {   discard_token_location(beginning_debug_location);
             memoryerror("MAX_LINESPACE", MAX_LINESPACE);
+        }
     }
     else {
         if (mark + 165 >= MAX_LINESPACE)
+        {   discard_token_location(beginning_debug_location);
             memoryerror("MAX_LINESPACE", MAX_LINESPACE);
+        }
     }
 
     Inform_verbs[verbnum].l[line] = mark;
@@ -419,7 +459,8 @@ into Inform, so suggest rewriting grammar using general parsing routines");
     {   get_next_token();
         bytecode = 0; wordcode = 0;
         if ((token_type == SEP_TT) && (token_value == SEMICOLON_SEP))
-        {   ebf_error("'->' clause", token_text);
+        {   discard_token_location(beginning_debug_location);
+            ebf_error("'->' clause", token_text);
             return FALSE;
         }
         if ((token_type == SEP_TT) && (token_value == ARROW_SEP))
@@ -462,7 +503,8 @@ into Inform, so suggest rewriting grammar using general parsing routines");
                      get_next_token();
                      if ((token_type != SYMBOL_TT)
                          || (stypes[token_value] != ROUTINE_T))
-                     {   ebf_error("routine name after 'noun='", token_text);
+                     {   discard_token_location(beginning_debug_location);
+                         ebf_error("routine name after 'noun='", token_text);
                          panic_mode_error_recovery();
                          return FALSE;
                      }
@@ -516,7 +558,8 @@ are using Library 6/3 or later");
 
                  get_next_token();
                  if (!((token_type==SEP_TT)&&(token_value==SETEQUALS_SEP)))
-                 {   ebf_error("'=' after 'scope'", token_text);
+                 {   discard_token_location(beginning_debug_location);
+                     ebf_error("'=' after 'scope'", token_text);
                      panic_mode_error_recovery();
                      return FALSE;
                  }
@@ -524,7 +567,8 @@ are using Library 6/3 or later");
                  get_next_token();
                  if ((token_type != SYMBOL_TT)
                      || (stypes[token_value] != ROUTINE_T))
-                 {   ebf_error("routine name after 'scope='", token_text);
+                 {   discard_token_location(beginning_debug_location);
+                     ebf_error("routine name after 'scope='", token_text);
                      panic_mode_error_recovery();
                      return FALSE;
                  }
@@ -536,7 +580,8 @@ are using Library 6/3 or later");
                  sflags[token_value] |= USED_SFLAG;
              }
         else if ((token_type == SEP_TT) && (token_value == SETEQUALS_SEP))
-             {   error("'=' is only legal here as 'noun=Routine'");
+             {   discard_token_location(beginning_debug_location);
+                 error("'=' is only legal here as 'noun=Routine'");
                  panic_mode_error_recovery();
                  return FALSE;
              }
@@ -545,8 +590,10 @@ are using Library 6/3 or later");
                  if ((token_type != SYMBOL_TT)
                      || ((stypes[token_value] != ATTRIBUTE_T)
                          && (stypes[token_value] != ROUTINE_T)))
-                 {   error_named("No such grammar token as", token_text);
-                     panic_mode_error_recovery(); return FALSE;
+                 {   discard_token_location(beginning_debug_location);
+                     error_named("No such grammar token as", token_text);
+                     panic_mode_error_recovery();
+                     return FALSE;
                  }
                  if (stypes[token_value]==ATTRIBUTE_T)
                  {   if (grammar_version_number == 1)
@@ -597,7 +644,8 @@ tokens in any line (unless you're compiling with library 6/3 or later)");
     dont_enter_into_symbol_table = FALSE;
 
     if (token_type != DQ_TT)
-    {   ebf_error("name of new or existing action", token_text);
+    {   discard_token_location(beginning_debug_location);
+        ebf_error("name of new or existing action", token_text);
         panic_mode_error_recovery();
         return FALSE;
     }
@@ -620,6 +668,20 @@ Library 6/3 or later");
 
     mark = Inform_verbs[verbnum].l[line];
 
+    if (debugfile_switch)
+    {   debug_file_printf("<table-entry>");
+        debug_file_printf("<type>grammar line</type>");
+        debug_file_printf("<address>");
+        write_debug_grammar_backpatch(mark);
+        debug_file_printf("</address>");
+        debug_file_printf("<end-address>");
+        write_debug_grammar_backpatch(grammar_lines_top);
+        debug_file_printf("</end-address>");
+        write_debug_locations
+            (get_token_location_end(beginning_debug_location));
+        debug_file_printf("</table-entry>");
+    }
+
     if (!glulx_mode) {
         if (reverse_action)
             j = j + 0x400;
@@ -688,7 +750,7 @@ extern void make_verb(void)
     for (i=0; i<no_given; i++)
     {   dictionary_add(English_verbs_given[i],
             0x41 + ((meta_verb_flag)?0x02:0x00),
-            0xff-Inform_verb, 0);
+            (glulx_mode)?(0xffff-Inform_verb):(0xff-Inform_verb), 0);
         register_verb(English_verbs_given[i], Inform_verb);
     }
 
@@ -720,7 +782,7 @@ extern void extend_verb(void)
 {
     /*  Parse an entire Extend ... directive.                                */
 
-    int Inform_verb, k, l, lines, extend_mode;
+    int Inform_verb = -1, k, l, lines, extend_mode;
 
     directive_keywords.enabled = TRUE;
     directives.enabled = FALSE;
@@ -737,7 +799,8 @@ extern void extend_verb(void)
             if ((l!=-1) && (Inform_verb!=l))
               warning_named("Verb disagrees with previous verbs:", token_text);
             l = Inform_verb;
-            dictionary_set_verb_number(token_text, 0xff-no_Inform_verbs);
+            dictionary_set_verb_number(token_text,
+              (glulx_mode)?(0xffff-no_Inform_verbs):(0xff-no_Inform_verbs));
             /* make call to renumber verb in English_verb_list too */
             if (find_or_renumber_verb(token_text, &no_Inform_verbs) == -1)
               warning_named("Verb to extend not found in English_verb_list:",
@@ -845,7 +908,7 @@ extern void verbs_allocate_arrays(void)
                                 "grammar token routines");
     adjectives            = my_calloc(sizeof(int32),   MAX_ADJECTIVES,
                                 "adjectives");
-    adjective_sort_code   = my_calloc(DICT_WORD_SIZE,  MAX_ADJECTIVES,
+    adjective_sort_code   = my_calloc(DICT_WORD_BYTES, MAX_ADJECTIVES,
                                 "adjective sort codes");
 
     English_verb_list     = my_malloc(MAX_VERBSPACE, "register of verbs");
